#include "tmEventSetup/esObjectHandle.hh"

#include "tmEventSetup/tmEventSetup.hh"
#include "tmEventSetup/esCutHandle.hh"
#include "tmUtil/tmUtil.hh"

#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/replace.hpp>

namespace tmeventsetup
{

esObjectHandle::esObjectHandle(const Object::Item& item)
{
  init(item);
}


esObjectHandle::esObjectHandle(const Object::Item& item, const tmtable::Table& cuts_in_algo)
{
  init(item);
  for (const auto& item_cut: item.cuts)
  {
    for (const auto& algo_cut: cuts_in_algo)
    {
      tmtable::Row& cut = const_cast<tmtable::Row&>(algo_cut);
      if (item_cut != cut["name"]) continue;

      // add object name if it is not included in the cut["type"]
      if (cut["type"].find("-") == std::string::npos)
      {
        cut["type"] = item.name + "-" + cut["type"];
      }
      esCutHandle handle = esCutHandle(cut);
      esCut& ref = handle;
      cuts_.push_back(ref);
    }
  }
}


void
esObjectHandle::init(const Object::Item& item)
{
  bool isThreshold = true;

  name_ = item.getObjectName();

  type_ = static_cast<esObjectType>(Undef);
  if (item.name == Object::MU) type_ = Muon;
  else if (item.name == Object::EG) type_ = Egamma;
  else if (item.name == Object::TAU) type_ = Tau;
  else if (item.name == Object::JET) type_ = Jet;
  else if (item.name == Object::ETT) type_ = ETT;
  else if (item.name == Object::HTT) type_ = HTT;
  else if (item.name == Object::ETM) type_ = ETM;
  else if (item.name == Object::HTM) type_ = HTM;
  else if (item.name == Object::MBT0HFP)
  {
    type_ = MBT0HFP;
    isThreshold = false;
  }
  else if (item.name == Object::MBT1HFP)
  {
    type_ = MBT1HFP;
    isThreshold = false;
  }
  else if (item.name == Object::MBT0HFM)
  {
    type_ = MBT0HFM;
    isThreshold = false;
  }
  else if (item.name == Object::MBT1HFM)
  {
    type_ = MBT1HFM;
    isThreshold = false;
  }
  else if (item.name == Object::TOWERCOUNT)
  {
    type_ = TOWERCOUNT;
    isThreshold = false;
  }
  else if (item.name == Object::ASYMET)
  {
    type_ = ASYMET;
    isThreshold = false;
  }
  else if (item.name == Object::ASYMHT)
  {
    type_ = ASYMHT;
    isThreshold = false;
  }
  else if (item.name == Object::ASYMETHF)
  {
    type_ = ASYMETHF;
    isThreshold = false;
  }
  else if (item.name == Object::ASYMHTHF)
  {
    type_ = ASYMHTHF;
    isThreshold = false;
  }
  else if (item.name == Object::CENT0)
  {
    type_ = CENT0;
    isThreshold = false;
  }
  else if (item.name == Object::CENT1)
  {
    type_ = CENT1;
    isThreshold = false;
  }
  else if (item.name == Object::CENT2)
  {
    type_ = CENT2;
    isThreshold = false;
  }
  else if (item.name == Object::CENT3)
  {
    type_ = CENT3;
    isThreshold = false;
  }
  else if (item.name == Object::CENT4)
  {
    type_ = CENT4;
    isThreshold = false;
  }
  else if (item.name == Object::CENT5)
  {
    type_ = CENT5;
    isThreshold = false;
  }
  else if (item.name == Object::CENT6)
  {
    type_ = CENT6;
    isThreshold = false;
  }
  else if (item.name == Object::CENT7)
  {
    type_ = CENT7;
    isThreshold = false;
  }
  else if (item.name == Object::MUS0)
  {
    type_ = MUS0;
    isThreshold = false;
  }
  else if (item.name == Object::MUS1)
  {
    type_ = MUS1;
    isThreshold = false;
  }
  else if (item.name == Object::MUSOOT0)
  {
    type_ = MUSOOT0;
    isThreshold = false;
  }
  else if (item.name == Object::MUSOOT1)
  {
    type_ = MUSOOT1;
    isThreshold = false;
  }
  else if (item.name == Object::ETTEM) type_ = ETTEM;
  else if (item.name == Object::ETMHF) type_ = ETMHF;
  else if (item.name.rfind(Object::EXT, 0) == 0) type_ = EXT;
  else
  {
    TM_FATAL_ERROR("unknown type: " << TM_QUOTE(item.name));
  }

  // If object is a signal
  if (Object::signalName.count(item.name))
  {
    bx_offset_ = boost::lexical_cast<int>(item.bx_offset);
    return;
  }

  if (type_ == EXT)
  {
    bx_offset_ = boost::lexical_cast<int>(item.bx_offset);
    ext_signal_name_ = getExternalName(name_);
    return;
  }

  comparison_operator_ = static_cast<esComparisonOperator>(Undef);
  if (item.comparison == Object::GE) comparison_operator_ = GE;
  else if (item.comparison == Object::NE) comparison_operator_ = NE;
  else if (item.comparison == Object::EQ) comparison_operator_ = EQ;
  else
  {
    TM_FATAL_ERROR("unknown comparison operator " << TM_QUOTE(item.comparison));
  }

  bx_offset_ = boost::lexical_cast<int>(item.bx_offset);
  std::string threshold = item.threshold;
  // TODO: implement common threshold conversion...
  boost::algorithm::replace_all(threshold, "p", ".");
  threshold_ = boost::lexical_cast<double>(threshold);

  // Set proper cut type for object.
  std::string cut_type = ET_THR;
  if (not isThreshold) cut_type = COUNT;

  // Create table for threshold cut.
  tmtable::Row cut;
  cut["name"] = item.name + "-" + cut_type + "_" + item.threshold; // MU-ET_1p2
  cut["object"] = item.name;
  cut["type"] = item.name + "-" + cut_type;
  cut["minimum"] = boost::str(boost::format("%+23.16E") % threshold_);
  cut["maximum"] = "";
  cut["data"] = "";

  // Create threshold cut from table.
  esCutHandle handle = esCutHandle(cut);
  esCut& ref = handle;
  cuts_.push_back(ref);
}


void
esObjectHandle::setExternalChannelId(const std::map<std::string, unsigned int>& map)
{
  if (type_ == EXT)
  {
    if (map.count(ext_signal_name_))
    {
      ext_channel_id_ = map.at(ext_signal_name_);
    }
    else
    {
      TM_FATAL_ERROR("unknown external signal name " << TM_QUOTE(ext_signal_name_));
    }
  }
}


void
esObjectHandle::print() const
{
  print(*this);
}


void
esObjectHandle::print(const esObject& object)
{
  TM_LOG_INF("");
  std::cout << "  name: " << object.getName() << "\n";
  std::cout << "  type: " << object.getType() << "\n";
  std::cout << "  comparison_operator: " << object.getComparisonOperator() << "\n";
  std::cout << "  bx_offset: " << object.getBxOffset() << "\n";
  std::cout << "  threshold: " << object.getThreshold() << "\n";
  const std::vector<esCut>& cuts = object.getCuts();
  std::cout << "  cuts_size: " << cuts.size() << "\n";
  for (const auto& cut: cuts) esCutHandle::print(cut);
}

} // namespace tmeventsetup
