#include "tmEventSetup/esScaleHandle.hh"
#include "tmGrammar/Cut.hh"
#include "tmGrammar/Object.hh"
#include "tmUtil/tmUtil.hh"

#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string.hpp>

#include <algorithm>

namespace tmeventsetup
{

esScaleHandle::esScaleHandle(const tmtable::Row& scale)
{
  object_ = static_cast<esObjectType>(Undef);
  const std::string obj = scale.at("object");
  if (obj == Object::MU) object_ = Muon;
  else if (obj == Object::EG) object_ = Egamma;
  else if (obj == Object::TAU) object_ = Tau;
  else if (obj == Object::JET) object_ = Jet;
  else if (obj == Object::HTT) object_ = HTT;
  else if (obj == Object::ETT) object_ = ETT;
  else if (obj == Object::ETM) object_ = ETM;
  else if (obj == Object::HTM) object_ = HTM;
  else if (obj == Object::MBT0HFP) object_ = MBT0HFP;
  else if (obj == Object::MBT1HFP) object_ = MBT1HFP;
  else if (obj == Object::MBT0HFM) object_ = MBT0HFM;
  else if (obj == Object::MBT1HFM) object_ = MBT1HFM;
  else if (obj == Object::ETTEM) object_ = ETTEM;
  else if (obj == Object::ETMHF) object_ = ETMHF;
  else if (obj == Object::TOWERCOUNT) object_ = TOWERCOUNT;
  else if (obj == Object::ASYMET) object_ = ASYMET;
  else if (obj == Object::ASYMHT) object_ = ASYMHT;
  else if (obj == Object::ASYMETHF) object_ = ASYMETHF;
  else if (obj == Object::ASYMHTHF) object_ = ASYMHTHF;
  else if (obj == PRECISION) object_ = Precision;
  else
  {
    TM_FATAL_ERROR("unknown object: " << TM_QUOTE(obj));
  }

  type_ = static_cast<esScaleType>(Undef);
  const std::string type = scale.at("type");
  if (type == ET_THR) type_ = EtScale;
  else if (type == Cut::UPT) type_ = UnconstrainedPtScale;
  else if (type == Cut::ETA) type_ = EtaScale;
  else if (type == Cut::PHI) type_ = PhiScale;
  else if (type == COUNT) type_ = CountScale;
  else if (boost::algorithm::ends_with(type, "-" + PRECISION_DELTA)) type_ = DeltaPrecision;
  else if (boost::algorithm::ends_with(type, "-" + PRECISION_OVRM_DELTA)) type_ = OvRmDeltaPrecision;
  else if (boost::algorithm::ends_with(type, "-" + PRECISION_MASS)) type_ = MassPrecision;
  else if (boost::algorithm::ends_with(type, "-" + PRECISION_MASSPT)) type_ = MassPtPrecision;
  else if (boost::algorithm::ends_with(type, "-" + PRECISION_MATH)) type_ = MathPrecision;
  else if (boost::algorithm::ends_with(type, "-" + PRECISION_TBPT)) type_ = TwoBodyPtPrecision;
  else if (boost::algorithm::ends_with(type, "-" + PRECISION_TBPT_MATH)) type_ = TwoBodyPtMathPrecision;
  else if (boost::algorithm::ends_with(type, "-" + PRECISION_INVERSE_DR)) type_ = InverseDeltaRPrecision;
  else
  {
    TM_FATAL_ERROR("unknown type " << TM_QUOTE(type));
  }

  name_ =  obj + "-" + type;
  minimum_ = boost::lexical_cast<double>(scale.at("minimum"));
  maximum_ = boost::lexical_cast<double>(scale.at("maximum"));
  step_ = boost::lexical_cast<double>(scale.at("step"));
  n_bits_ = boost::lexical_cast<unsigned int>(scale.at("n_bits"));
}


void
esScaleHandle::addBin(const esBin& bin)
{
  bins_.push_back(bin);
}


void
esScaleHandle::print() const
{
  print(*this);
}


void
esScaleHandle::print(const esScale& scale)
{
  TM_LOG_INF("");
  std::cout << "  name: " << scale.getName() << "\n";
  std::cout << "  object: " << scale.getObjectType() << "\n";
  std::cout << "  type: " << scale.getScaleType() << "\n";
  std::cout << "  minimum: " << scale.getMinimum() << "\n";
  std::cout << "  maximum: " << scale.getMaximum() << "\n";
  std::cout << "  step: " << scale.getStep() << "\n";
  std::cout << "  n_bits: " << scale.getNbits() << "\n";
  std::cout << "  bins_size: " << scale.getBins().size() << "\n";
  // for (size_t ii = 0; ii < bins_.size(); ii++) bins_.at(ii)->print();
}


void
esScaleHandle::sortBins()
{
  switch (type_)
  {
    // Sort eta scales by minimum to preserve two's complement hw_index
    case EtaScale:
      std::sort(bins_.begin(), bins_.end(), [](const esBin& a, const esBin& b) {
        return (a.minimum < b.minimum);
      });
      break;
    default:
      std::sort(bins_.begin(), bins_.end(), [](const esBin& a, const esBin& b) {
        return (a.hw_index < b.hw_index);
      });
      break;
  }
}

} // namespace tmeventsetup
